from collections.abc import Iterable, Mapping, Sequence
from typing import AbstractSet, Any, Optional  # noqa: UP035

import dagster._check as check
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.repository_definition import SINGLETON_REPOSITORY_NAME
from dagster._record import IHaveNew, record, record_custom
from dagster._serdes import create_snapshot_id, whitelist_for_serdes


@record_custom
class JobSubsetSelector(IHaveNew):
    """The information needed to resolve a job within a host process."""

    location_name: str
    repository_name: str
    job_name: str
    op_selection: Optional[Sequence[str]]
    asset_selection: Optional[AbstractSet[AssetKey]]
    asset_check_selection: Optional[AbstractSet[AssetCheckKey]]

    def __new__(
        cls,
        location_name: str,
        repository_name: str,
        job_name: str,
        op_selection: Optional[Sequence[str]],
        asset_selection: Optional[Iterable[AssetKey]] = None,
        asset_check_selection: Optional[Iterable[AssetCheckKey]] = None,
    ):
        # coerce iterables to sets
        asset_selection = frozenset(asset_selection) if asset_selection else None
        asset_check_selection = (
            frozenset(asset_check_selection) if asset_check_selection is not None else None
        )
        return super().__new__(
            cls,
            location_name=location_name,
            repository_name=repository_name,
            job_name=job_name,
            op_selection=op_selection,
            asset_selection=asset_selection,
            asset_check_selection=asset_check_selection,
        )

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "pipelineName": self.job_name,
            "solidSelection": self.op_selection,
        }

    @property
    def is_subset_selection(self) -> bool:
        return bool(self.op_selection or self.asset_selection or self.asset_check_selection)

    def with_op_selection(self, op_selection: Optional[Sequence[str]]) -> "JobSubsetSelector":
        check.invariant(
            self.op_selection is None,
            f"Can not invoke with_op_selection when op_selection={self.op_selection} is"
            " already set",
        )
        return JobSubsetSelector(
            self.location_name, self.repository_name, self.job_name, op_selection
        )

    @property
    def repository_selector(self) -> "RepositorySelector":
        return RepositorySelector(
            location_name=self.location_name,
            repository_name=self.repository_name,
        )


@whitelist_for_serdes
@record_custom
class JobSelector(IHaveNew):
    location_name: str
    repository_name: str
    job_name: str

    def __new__(
        cls,
        location_name: str,
        repository_name: Optional[str] = None,
        job_name: Optional[str] = None,
    ):
        check.invariant(
            job_name is not None,
            "Must provide job_name argument even though it is marked as optional in the "
            "function signature. repository_name, a truly optional parameter, is before "
            "that argument and actually optional. Use of keyword arguments is "
            "recommended to avoid confusion.",
        )

        return super().__new__(
            cls,
            location_name=location_name,
            repository_name=repository_name or SINGLETON_REPOSITORY_NAME,
            job_name=job_name,
        )

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "jobName": self.job_name,
        }

    @property
    def selector_id(self):
        return create_snapshot_id(self)

    @staticmethod
    def from_graphql_input(graphql_data):
        return JobSelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
            job_name=graphql_data["jobName"],
        )

    @property
    def repository_selector(self) -> "RepositorySelector":
        return RepositorySelector(
            location_name=self.location_name,
            repository_name=self.repository_name,
        )


@whitelist_for_serdes
@record
class RepositorySelector:
    location_name: str
    repository_name: str

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
        }

    @property
    def selector_id(self):
        return create_snapshot_id(self)

    @staticmethod
    def from_graphql_input(graphql_data):
        return RepositorySelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
        )


@record
class AssetGroupSelector:
    group_name: str
    location_name: str
    repository_name: str

    @staticmethod
    def from_graphql_input(graphql_data):
        return AssetGroupSelector(
            group_name=graphql_data["groupName"],
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
        )


@record(kw_only=False)
class CodeLocationSelector:
    location_name: str

    def to_repository_selector(self) -> RepositorySelector:
        return RepositorySelector(
            location_name=self.location_name,
            repository_name=SINGLETON_REPOSITORY_NAME,
        )


@record
class ScheduleSelector:
    location_name: str
    repository_name: str
    schedule_name: str

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "scheduleName": self.schedule_name,
        }

    @property
    def instigator_name(self) -> str:
        return self.schedule_name

    @staticmethod
    def from_graphql_input(graphql_data):
        return ScheduleSelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
            schedule_name=graphql_data["scheduleName"],
        )


@record
class ResourceSelector:
    location_name: str
    repository_name: str
    resource_name: str

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "resourceName": self.resource_name,
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        return ResourceSelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
            resource_name=graphql_data["resourceName"],
        )


@record
class SensorSelector:
    location_name: str
    repository_name: str
    sensor_name: str

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "sensorName": self.sensor_name,
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        return SensorSelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
            sensor_name=graphql_data["sensorName"],
        )

    @property
    def instigator_name(self) -> str:
        return self.sensor_name


@whitelist_for_serdes
@record
class InstigatorSelector:
    location_name: str
    repository_name: str
    name: str

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "name": self.name,
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        return InstigatorSelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
            name=graphql_data["name"],
        )

    def get_id(self) -> str:
        return create_snapshot_id(self)


@record
class GraphSelector:
    """The information needed to resolve a graph within a host process."""

    location_name: str
    repository_name: str
    graph_name: str

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "graphName": self.graph_name,
        }


@whitelist_for_serdes
@record
class PartitionSetSelector:
    """The information needed to resolve a partition set within a host process."""

    location_name: str
    repository_name: str
    partition_set_name: str

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "partitionSetName": self.partition_set_name,
        }


@record(kw_only=False)
class PartitionRangeSelector:
    """The information needed to resolve a partition range."""

    start: str
    end: str

    def to_graphql_input(self):
        return {
            "start": self.start,
            "end": self.end,
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        return PartitionRangeSelector(
            start=graphql_data["start"],
            end=graphql_data["end"],
        )


@record(kw_only=False)
class PartitionsSelector:
    """The information needed to define selection partitions."""

    ranges: Sequence[PartitionRangeSelector]

    def to_graphql_input(self):
        return {"ranges": [partition_range.to_graphql_input() for partition_range in self.ranges]}

    @staticmethod
    def from_graphql_input(graphql_data: Mapping[str, Any]) -> "PartitionsSelector":
        if "ranges" in graphql_data:
            check.invariant(
                "range" not in graphql_data,
                "Received partitionsSelector with values for both 'range' and 'ranges'. Only one should be provided.",
            )
            return PartitionsSelector(
                ranges=[
                    PartitionRangeSelector.from_graphql_input(range_data)
                    for range_data in graphql_data["ranges"]
                ]
            )

        if "range" in graphql_data:
            return PartitionsSelector(
                ranges=[PartitionRangeSelector.from_graphql_input(graphql_data["range"])]
            )

        check.failed(
            "Received partitionsSelector without values for either 'range' or 'ranges'. One should be provided.",
        )


@record
class PartitionsByAssetSelector:
    """The information needed to define partitions selection for a given asset key."""

    asset_key: AssetKey
    partitions: Optional[PartitionsSelector] = None

    def to_graphql_input(self):
        return {
            "assetKey": self.asset_key.to_graphql_input(),
            "partitions": self.partitions.to_graphql_input() if self.partitions else None,
        }

    @staticmethod
    def from_graphql_input(graphql_data) -> "PartitionsByAssetSelector":
        asset_key = graphql_data["assetKey"]
        partitions = graphql_data.get("partitions")
        return PartitionsByAssetSelector(
            asset_key=AssetKey.from_graphql_input(asset_key),
            partitions=PartitionsSelector.from_graphql_input(partitions) if partitions else None,
        )



import os
import random
import re
import string
import uuid
import warnings
from collections import OrderedDict, deque
from collections.abc import Iterable, Iterator, Mapping, Sequence
from concurrent.futures import Future, ThreadPoolExecutor, TimeoutError
from contextvars import copy_context
from typing import (  # noqa: UP035
    AbstractSet,
    Any,
    Callable,
    Final,
    Optional,
    TypedDict,
    TypeVar,
    Union,
    cast,
)
from weakref import WeakSet

import toposort as toposort_

import dagster._check as check
from dagster._utils import library_version_from_core_version, parse_package_version

BACKFILL_TAG_LENGTH = 8

PYTHON_LOGGING_LEVELS_MAPPING: Final[Mapping[str, int]] = OrderedDict(
    {"CRITICAL": 50, "ERROR": 40, "WARNING": 30, "INFO": 20, "DEBUG": 10}
)

PYTHON_LOGGING_LEVELS_ALIASES: Final[Mapping[str, str]] = OrderedDict(
    {"FATAL": "CRITICAL", "WARN": "WARNING"}
)

PYTHON_LOGGING_LEVELS_NAMES = frozenset(
    [
        level_name.lower()
        for level_name in sorted(
            list(PYTHON_LOGGING_LEVELS_MAPPING.keys()) + list(PYTHON_LOGGING_LEVELS_ALIASES.keys())
        )
    ]
)

T = TypeVar("T", bound=Any)


def coerce_valid_log_level(log_level: Union[str, int]) -> int:
    """Convert a log level into an integer for consumption by the low-level Python logging API."""
    if isinstance(log_level, int):
        return log_level
    str_log_level = check.str_param(log_level, "log_level")
    check.invariant(
        str_log_level.lower() in PYTHON_LOGGING_LEVELS_NAMES,
        "Bad value for log level {level}: permissible values are {levels}.".format(
            level=str_log_level,
            levels=", ".join(
                [f"'{level_name.upper()}'" for level_name in PYTHON_LOGGING_LEVELS_NAMES]
            ),
        ),
    )
    str_log_level = PYTHON_LOGGING_LEVELS_ALIASES.get(log_level.upper(), log_level.upper())
    return PYTHON_LOGGING_LEVELS_MAPPING[str_log_level]


def toposort(
    data: Mapping[T, AbstractSet[T]], sort_key: Optional[Callable[[T], Any]] = None
) -> Sequence[Sequence[T]]:
    # Workaround a bug in older versions of toposort that choke on frozenset
    data = {k: set(v) if isinstance(v, frozenset) else v for k, v in data.items()}
    return [sorted(list(level), key=sort_key) for level in toposort_.toposort(data)]


def toposort_flatten(data: Mapping[T, AbstractSet[T]]) -> Sequence[T]:
    return [item for level in toposort(data) for item in level]


def make_new_run_id() -> str:
    return str(uuid.uuid4())


def is_valid_run_id(run_id: str):
    try:
        uuid.UUID(run_id, version=4)
        return True
    except ValueError:
        return False


def make_new_backfill_id() -> str:
    return "".join(random.choice(string.ascii_lowercase) for x in range(BACKFILL_TAG_LENGTH))


def str_format_list(items: Iterable[object]) -> str:
    return "[{items}]".format(items=", ".join([f"'{item}'" for item in items]))


def str_format_set(items: Iterable[object]) -> str:
    return "[{items}]".format(items=", ".join([f"'{item}'" for item in items]))


def check_dagster_package_version(library_name: str, library_version: str) -> None:
    # This import must be internal in order for this function to be testable
    from dagster.version import __version__

    parsed_lib_version = parse_package_version(library_version)
    if parsed_lib_version.release[0] >= 1:
        if library_version != __version__:
            message = (
                f"Found version mismatch between `dagster` ({__version__})"
                f"and `{library_name}` ({library_version})"
            )
            warnings.warn(message)
    else:
        target_version = library_version_from_core_version(__version__)
        if library_version != target_version:
            message = (
                f"Found version mismatch between `dagster` ({__version__}) "
                f"expected library version ({target_version}) "
                f"and `{library_name}` ({library_version})."
            )
            warnings.warn(message)


def get_env_var_name(env_var_str: str):
    if "=" in env_var_str:
        return env_var_str.split("=", maxsplit=1)[0]
    else:
        return env_var_str


def parse_env_var(env_var_str: str) -> tuple[str, str]:
    if "=" in env_var_str:
        split = env_var_str.split("=", maxsplit=1)
        return (split[0], split[1])
    else:
        env_var_value = os.getenv(env_var_str)
        if env_var_value is None:
            raise Exception(f"Tried to load environment variable {env_var_str}, but it was not set")
        return (env_var_str, cast(str, env_var_value))


class RequestUtilizationMetrics(TypedDict):
    """A dict of utilization metrics for a threadpool executor. We use generic language in case we use this metrics in a scenario where we switch away from a threadpool executor at a later time."""

    max_concurrent_requests: Optional[int]
    num_running_requests: Optional[int]
    num_queued_requests: Optional[int]


class FuturesAwareThreadPoolExecutor(ThreadPoolExecutor):
    def __init__(
        self,
        max_workers: Optional[int] = None,
        thread_name_prefix: str = "",
        initializer: Any = None,
        initargs: tuple[Any, ...] = (),
    ) -> None:
        super().__init__(max_workers, thread_name_prefix, initializer, initargs)
        # The default threadpool class doesn't track the futures it creates,
        # so if we want to be able to count the number of running futures, we need to do it ourselves.
        self._tracked_futures: WeakSet[Future] = WeakSet()

    def submit(self, fn, *args, **kwargs) -> Future:
        new_future = super().submit(fn, *args, **kwargs)
        self._tracked_futures.add(new_future)
        return new_future

    @property
    def max_workers(self) -> int:
        return self._max_workers

    @property
    def weak_tracked_futures_count(self) -> int:
        return len(self._tracked_futures)

    @property
    def num_running_futures(self) -> int:
        return sum(1 for f in self._tracked_futures if not f.done()) - self.num_queued_futures

    @property
    def num_queued_futures(self) -> int:
        return self._work_queue.qsize()

    def get_current_utilization_metrics(self) -> RequestUtilizationMetrics:
        return {
            "max_concurrent_requests": self.max_workers,
            "num_running_requests": self.num_running_futures,
            "num_queued_requests": self.num_queued_futures,
        }


class InheritContextThreadPoolExecutor(FuturesAwareThreadPoolExecutor):
    """A ThreadPoolExecutor that copies over contextvars at submit time."""

    def submit(self, fn, *args, **kwargs):
        ctx = copy_context()
        return super().submit(ctx.run, fn, *args, **kwargs)


def is_valid_email(email: str) -> bool:
    regex = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b"
    return bool(re.fullmatch(regex, email))


T = TypeVar("T")
P = TypeVar("P")


def imap(
    executor: ThreadPoolExecutor,
    iterable: Iterator[T],
    func: Callable[[T], P],
) -> Iterator[P]:
    """A version of `concurrent.futures.ThreadpoolExecutor.map` which tails the input iterator in
    a separate thread. This means that the map function can begin processing and yielding results from
    the first elements of the iterator before the iterator is fully consumed.

    Args:
        executor: The ThreadPoolExecutor to use for parallel execution.
        iterable: The iterator to apply the function to.
        func: The function to apply to each element of the iterator.
    """
    work_queue: deque[Future] = deque([])

    # create a small task which waits on the iterator
    # and enqueues work items as they become available
    def _apply_func_to_iterator_results(iterable: Iterator) -> None:
        for arg in iterable:
            work_queue.append(executor.submit(func, arg))

    enqueuing_task = executor.submit(
        _apply_func_to_iterator_results,
        iterable,
    )

    while True:
        if (not enqueuing_task or enqueuing_task.done()) and len(work_queue) == 0:
            break

        if len(work_queue) > 0:
            current_work_item = work_queue[0]

            try:
                yield current_work_item.result(timeout=0.1)
                work_queue.popleft()
            except TimeoutError:
                pass

    # Ensure any exceptions from the enqueuing task processing the iterator are raised,
    # after all work items have been processed.
    exc = enqueuing_task.exception()
    if exc:
        raise exc


def exhaust_iterator_and_yield_results_with_exception(iterable: Iterator[T]) -> Iterator[T]:
    """Fully exhausts an iterator and then yield its results. If the iterator raises an exception,
    raise that exception at the position in the iterator where it was originally raised.

    This is useful if we want to exhaust an iterator, but don't want to discard the elements
    that were yielded before the exception was raised.
    """
    results = []
    caught_exception = None
    try:
        for result in iterable:
            results.append(result)
    except Exception as e:
        caught_exception = e

    yield from results
    if caught_exception:
        raise caught_exception



import time
from datetime import datetime, timedelta, timezone, tzinfo
from typing import Union

import dagster._check as check
from dagster._vendored.dateutil import parser

try:
    # zoneinfo is python >= 3.9
    from zoneinfo import ZoneInfo as _timezone_from_string
except:
    from dagster._vendored.dateutil.tz import gettz as _timezone_from_string


def _mockable_get_current_datetime() -> datetime:
    # Can be mocked in tests by freeze_time()
    return datetime.now(tz=timezone.utc)


def get_current_datetime(tz="UTC") -> datetime:
    """Return the current datetime. Will always have a timezone
    (defaults to UTC if none is specified). Value can be mocked
    via dagster._core.test_utils.freeze_time.
    """
    utc_datetime = _mockable_get_current_datetime()

    if tz == "utc" or tz == "UTC":
        return utc_datetime

    return utc_datetime.astimezone(get_timezone(tz))


def _mockable_get_current_timestamp() -> float:
    return time.time()


def get_current_datetime_midnight(tz="UTC") -> datetime:
    """Return the current date at midnight as a datetime object. Will always have a timezone
    (defaults to UTC if none is specified). Value can be mocked via dagster._core.test_utils.freeze_time.
    """
    return get_current_datetime(tz).replace(hour=0, minute=0, second=0, microsecond=0)


def get_current_timestamp() -> float:
    """Return the current unix timestamp. Value can be mocked
    via dagster._core.test_utils.freeze_time.
    """
    # Like time.time() but can be mocked in tests by freeze_time()
    return _mockable_get_current_timestamp()


def get_timezone(timezone_name: str) -> tzinfo:
    """Creates a tzinfo object with the given IANA timezone name."""
    if timezone_name == "utc" or timezone_name == "UTC":
        return timezone.utc

    return check.not_none(_timezone_from_string(timezone_name))


def create_datetime(year, month, day, *args, **kwargs):
    """Creates a datetime object. Same arguments as datetime.datetime
    constructor, but will always have a timezone (defaults to UTC if none is specified).
    """
    tz = kwargs.pop("tz", "UTC")
    if isinstance(tz, str):
        tz = get_timezone(tz)
    return datetime(year, month, day, *args, **kwargs, tzinfo=tz)


def datetime_from_timestamp(timestamp: float, tz: Union[str, tzinfo] = timezone.utc) -> datetime:
    """Creates a datetime object from a unix timestamp. Will always have a timezone
    (defaults to UTC if none is specified).
    """
    if not tz:
        tzinfo = timezone.utc
    elif isinstance(tz, str):
        tzinfo = get_timezone(tz)
    else:
        tzinfo = tz

    return datetime.fromtimestamp(timestamp, tz=tzinfo)


def utc_datetime_from_naive(dt: datetime) -> datetime:
    return dt.replace(tzinfo=timezone.utc)


def add_absolute_time(
    dt: datetime,
    *,
    hours=0,
    minutes=0,
    seconds=0,
    milliseconds=0,
    microseconds=0,
):
    """Behaves like adding a time using a timedelta, but handles fall DST transitions correctly
    without skipping an hour ahead.
    """
    return (
        dt.astimezone(timezone.utc)
        + timedelta(
            seconds=seconds,
            microseconds=microseconds,
            milliseconds=milliseconds,
            minutes=minutes,
            hours=hours,
        )
    ).astimezone(dt.tzinfo)


def parse_time_string(datetime_str) -> datetime:
    """Like dateutil.parser.parse, but always includes a timezone (defaults to UTC if
    no timezone is included in the timezone string).
    """
    dt = parser.parse(datetime_str)

    if not dt.tzinfo:  # pyright: ignore[reportAttributeAccessIssue]
        dt = dt.replace(tzinfo=timezone.utc)  # pyright: ignore[reportAttributeAccessIssue]

    return dt  # pyright: ignore[reportReturnType]



from collections.abc import Mapping, Sequence
from typing import Any, Final, NamedTuple, Optional

import dagster._check as check
from dagster._core.code_pointer import CodePointer
from dagster._serdes import create_snapshot_id, whitelist_for_serdes

DEFAULT_DAGSTER_ENTRY_POINT: Final = ["dagster"]


def get_python_environment_entry_point(executable_path: str) -> Sequence[str]:
    return [executable_path, "-m", "dagster"]


@whitelist_for_serdes
class RepositoryPythonOrigin(
    NamedTuple(
        "_RepositoryPythonOrigin",
        [
            ("executable_path", str),
            ("code_pointer", CodePointer),
            ("container_image", Optional[str]),
            ("entry_point", Optional[Sequence[str]]),
            ("container_context", Optional[Mapping[str, Any]]),
        ],
    ),
):
    """Args:
    executable_path (str): The Python executable of the user process.
    code_pointer (CodePoitner): Once the process has started, an object that can be used to
        find and load the repository in code.
    container_image (Optinonal[str]): The image to use when creating a new container that
        loads the repository. Only used in execution environments that start containers.
    entry_point (Optional[List[str]]): The entry point to use when starting a new process
        to load the repository. Defaults to ["dagster"] (and may differ from the executable_path).
    container_context (Optional[Dict[str, Any]]): Additional context to use when creating a new
        container that loads the repository. Keys can be specific to a given compute substrate
        (for example, "docker", "k8s", etc.).
    """

    def __new__(
        cls,
        executable_path: str,
        code_pointer: CodePointer,
        container_image: Optional[str] = None,
        entry_point: Optional[Sequence[str]] = None,
        container_context: Optional[Mapping[str, Any]] = None,
    ):
        return super().__new__(
            cls,
            check.str_param(executable_path, "executable_path"),
            check.inst_param(code_pointer, "code_pointer", CodePointer),
            check.opt_str_param(container_image, "container_image"),
            (
                check.sequence_param(entry_point, "entry_point", of_type=str)
                if entry_point is not None
                else None
            ),
            (
                check.opt_mapping_param(container_context, "container_context")
                if container_context is not None
                else None
            ),
        )

    def get_id(self) -> str:
        return create_snapshot_id(self)

    def get_job_origin(self, job_name: str) -> "JobPythonOrigin":
        check.str_param(job_name, "pipeline_name")
        return JobPythonOrigin(job_name, self)


@whitelist_for_serdes(
    storage_name="PipelinePythonOrigin",
    storage_field_names={"job_name": "pipeline_name"},
)
class JobPythonOrigin(
    NamedTuple(
        "_JobPythonOrigin",
        [
            ("job_name", str),
            ("repository_origin", RepositoryPythonOrigin),
        ],
    )
):
    def __new__(cls, job_name: str, repository_origin: RepositoryPythonOrigin):
        return super().__new__(
            cls,
            check.str_param(job_name, "job_name"),
            check.inst_param(repository_origin, "repository_origin", RepositoryPythonOrigin),
        )

    def get_id(self) -> str:
        return create_snapshot_id(self)

    @property
    def executable_path(self) -> str:
        return self.repository_origin.executable_path

    def get_repo_pointer(self) -> CodePointer:
        return self.repository_origin.code_pointer



from dagster._config.pythonic_config.config import (
    Config as Config,
    PermissiveConfig as PermissiveConfig,
    _config_value_to_dict_representation as _config_value_to_dict_representation,
    infer_schema_from_config_class as infer_schema_from_config_class,
)
from dagster._config.pythonic_config.conversion_utils import (
    IAttachDifferentObjectToOpContext as IAttachDifferentObjectToOpContext,
    infer_schema_from_config_annotation as infer_schema_from_config_annotation,
    safe_is_subclass as safe_is_subclass,
)
from dagster._config.pythonic_config.io_manager import (
    ConfigurableIOManager as ConfigurableIOManager,
    ConfigurableIOManagerFactory as ConfigurableIOManagerFactory,
    ConfigurableIOManagerFactoryResourceDefinition as ConfigurableIOManagerFactoryResourceDefinition,
    ConfigurableLegacyIOManagerAdapter as ConfigurableLegacyIOManagerAdapter,
)
from dagster._config.pythonic_config.resource import (
    ConfigurableResource as ConfigurableResource,
    ConfigurableResourceFactory as ConfigurableResourceFactory,
    ConfigurableResourceFactoryResourceDefinition as ConfigurableResourceFactoryResourceDefinition,
    PartialResource as PartialResource,
    ResourceDependency as ResourceDependency,
    is_coercible_to_resource as is_coercible_to_resource,
    validate_resource_annotated_function as validate_resource_annotated_function,
)



"""This subpackage contains all classes that host processes (e.g. dagster-webserver)
use to manipulate and represent definitions that are resident
in user processes and containers.  e.g. ExternalPipeline.

It also contains classes that represent historical representations
that have been persisted. e.g. HistoricalPipeline
"""

from dagster._core.remote_representation.external import (
    RemoteExecutionPlan as RemoteExecutionPlan,
    RemoteJob as RemoteJob,
    RemotePartitionSet as RemotePartitionSet,
    RemoteRepository as RemoteRepository,
    RemoteSchedule as RemoteSchedule,
    RemoteSensor as RemoteSensor,
)
from dagster._core.remote_representation.external_data import (
    ExecutionParamsErrorSnap as ExecutionParamsErrorSnap,
    ExecutionParamsSnap as ExecutionParamsSnap,
    JobDataSnap as JobDataSnap,
    JobRefSnap as JobRefSnap,
    PartitionConfigSnap as PartitionConfigSnap,
    PartitionExecutionErrorSnap as PartitionExecutionErrorSnap,
    PartitionNamesSnap as PartitionNamesSnap,
    PartitionSetExecutionParamSnap as PartitionSetExecutionParamSnap,
    PartitionSetSnap as PartitionSetSnap,
    PartitionTagsSnap as PartitionTagsSnap,
    PresetSnap as PresetSnap,
    RemoteJobSubsetResult as RemoteJobSubsetResult,
    RepositoryErrorSnap as RepositoryErrorSnap,
    RepositorySnap as RepositorySnap,
    ScheduleExecutionErrorSnap as ScheduleExecutionErrorSnap,
    ScheduleSnap as ScheduleSnap,
    SensorExecutionErrorSnap as SensorExecutionErrorSnap,
    SensorSnap as SensorSnap,
    TargetSnap as TargetSnap,
)
from dagster._core.remote_representation.handle import (
    JobHandle as JobHandle,
    RepositoryHandle as RepositoryHandle,
)
from dagster._core.remote_representation.historical import HistoricalJob as HistoricalJob
from dagster._core.remote_representation.origin import (
    IN_PROCESS_NAME as IN_PROCESS_NAME,
    CodeLocationOrigin as CodeLocationOrigin,
    GrpcServerCodeLocationOrigin as GrpcServerCodeLocationOrigin,
    InProcessCodeLocationOrigin as InProcessCodeLocationOrigin,
    ManagedGrpcPythonEnvCodeLocationOrigin as ManagedGrpcPythonEnvCodeLocationOrigin,
    RemoteInstigatorOrigin as RemoteInstigatorOrigin,
    RemoteJobOrigin as RemoteJobOrigin,
    RemoteRepositoryOrigin as RemoteRepositoryOrigin,
)

# ruff: isort: split
from dagster._core.remote_representation.code_location import (
    CodeLocation as CodeLocation,
    GrpcServerCodeLocation as GrpcServerCodeLocation,
    InProcessCodeLocation as InProcessCodeLocation,
)
from dagster._core.remote_representation.job_index import JobIndex as JobIndex
from dagster._core.remote_representation.represented import RepresentedJob as RepresentedJob



import logging

from dagster._utils.error import (
    ExceptionInfo,
    SerializableErrorInfo,
    serializable_error_info_from_exc_info,
)


class DaemonErrorCapture:
    @staticmethod
    def default_process_exception(
        exc_info: ExceptionInfo,
        logger: logging.Logger,
        log_message: str,
    ) -> SerializableErrorInfo:
        error_info = serializable_error_info_from_exc_info(exc_info)
        logger.exception(log_message)
        return error_info

    # global behavior for how to handle unexpected exceptions
    process_exception = default_process_exception



import inspect
from abc import abstractmethod

from dagster._core.asset_graph_view.entity_subset import EntitySubset
from dagster._core.definitions.asset_key import T_EntityKey
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationResult,
    BuiltinAutomationCondition,
)
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._record import record


@record
class SubsetAutomationCondition(BuiltinAutomationCondition[T_EntityKey]):
    """Base class for simple conditions which compute a simple subset of the asset graph."""

    @property
    def requires_cursor(self) -> bool:
        return False

    @abstractmethod
    def compute_subset(
        self, context: AutomationContext[T_EntityKey]
    ) -> EntitySubset[T_EntityKey]: ...

    async def evaluate(
        self, context: AutomationContext[T_EntityKey]
    ) -> AutomationResult[T_EntityKey]:
        # don't compute anything if there are no candidates
        if context.candidate_subset.is_empty:
            true_subset = context.get_empty_subset()
        elif inspect.iscoroutinefunction(self.compute_subset):
            true_subset = await self.compute_subset(context)
        else:
            true_subset = self.compute_subset(context)

        return AutomationResult(context, true_subset)
